#ifndef CUDA_SIRECON_H
#define CUDA_SIRECON_H


#include "cudaSireconImpl.h"  // Or only a subset of it is needed?
/*
The following definitions are needed from cudaSirecon.h:

  ReconParams, ImageParams, DriftParams, ReconData
*/
#endif
